import pytest
from .tree import Tree

class AncestryTestCase():
  def __init__(self, ancestry, expected):
    self.ancestry = ancestry
    self.expected = expected
  
  def test(self, inst):
    assert inst.ancestry_branch(self.ancestry, {}) == self.expected

class AncestryGraftTestCase():
  def __init__(self, ancestries, expected):
    self.ancestries = ancestries
    self.expected = expected
  
  def test(self, inst):
    tree = {}
    for ancestry in self.ancestries:
      tree = inst.graft(tree, ancestry, {})
    assert tree == self.expected

def test_ancestry_branch_with_one_layer():
  AncestryTestCase(
    ['organizations/1234', 'projects/a'],
    {'organizations/1234' : {'projects/a':{}}}
  ).test(Tree(""))

def test_ancestry_branch_with_two_layers():
  AncestryTestCase(
    ['organizations/1234', 'folders/1234', 'projects/a'],
    {'organizations/1234' : {'folders/1234': {'projects/a' : {}}}}
  ).test(Tree(""))

def test_ancestry_branch_with_three_layers():
  AncestryTestCase(
    ['organizations/1234', 'folders/1234', 'folders/5678', 'projects/a'],
    {'organizations/1234' : {'folders/1234': {'folders/5678' : {'projects/a': {}}}}}
  ).test(Tree(""))

def test_grafting_simple():
  ancestries = [
    ['organizations/1234', 'projects/a'],
    ['organizations/1234', 'projects/b']
  ]
  expected = {'organizations/1234': {'projects/a':{}, 'projects/b':{}}}
  AncestryGraftTestCase(ancestries, expected).test(Tree(""))

def test_grafting_two_layers():
  ancestries = [
    ['organizations/1234', 'projects/a'],
    ['organizations/1234', 'folders/123','projects/b']
  ]
  expected = {'organizations/1234': {'projects/a':{}, 'folders/123': {'projects/b':{}} }}
  AncestryGraftTestCase(ancestries, expected).test(Tree(""))



import argparse
import json
import sys
from colorama import init, Fore, Style
from .tree import Tree

LEAF      = "└── "
LEAF_PLUS = "├── "
BAR_SPACE = "│   "
SPACE     = "    "

class Cli:
  def __init__(self):
    init()
    parser = argparse.ArgumentParser(description='Print out a GCP org heirarchy', prog="gcptree")
    parser.add_argument('org_id', nargs=1, 
                        help='GCP Organization ID')
    parser.add_argument('--format', default="text",
                        help='Output format (json or text)')
    parser.add_argument('--full-resource', action='store_const',
                        const=True, default = False,
                        help='API-parsable nodes where org and folder resource names are not resolved, i.e org/123 instead of example.com')

    self.args = parser.parse_args()
  
  def build_tree(self):
    t = Tree(self.args.org_id[0], self.args.full_resource)
    if t.cache.is_empty():
      print('Fetching GCP Resources, this may take a while (these results will be cached for an hour in {})... '.format(t.cache.filename), file=sys.stderr)
    tree = t.build()
    return tree
  
  def is_project(self, obj):
    return len(obj) > 0 and 'projectId' in obj

  def walk(self, tree, prefix=""):
    nodes = sorted(tree.keys())
    for i, node in enumerate(nodes):
      lchar, schar = (LEAF, SPACE) if i == len(nodes) - 1 else (LEAF_PLUS, BAR_SPACE)
      if self.is_project(tree[node]):
        formatted = node
        if tree[node]['lifecycleState'] != 'ACTIVE':
          formatted = Style.DIM + node + Style.RESET_ALL
        print(prefix + lchar + formatted)
      else:
        print(prefix + lchar + Style.BRIGHT + Fore.BLUE + node + Style.RESET_ALL)
        self.walk(tree[node], prefix + schar)

  def print_tree(self, tree):
    org = list(tree.keys())[0]
    print(Fore.GREEN + Style.BRIGHT + org + Style.RESET_ALL)
    self.walk(tree[org])

  def run(self):
    tree = self.build_tree()
    if self.args.format == 'json':
      print(json.dumps(tree, sort_keys=True, indent=4))
    elif self.args.format == 'text':
      self.print_tree(tree)
    else:
      print("Unsupported format")


from googleapiclient import discovery
from deepmerge import merge_or_raise
import json
import os
from .cache import Cache

class Tree():
  # Disable all the no-member violations in this class
  # pylint: disable=no-member
  def __init__(self, org_id, full_resource=False):
    self.org = "organizations/{}".format(org_id)
    self.resolve = not full_resource
    # v1 is the only version that supports project ancestry
    self.crm_v1 = discovery.build('cloudresourcemanager', 'v1')
    # v2 only supports folder name resolution
    self.crm_v2 = discovery.build('cloudresourcemanager', 'v2')
    self.cache = Cache()

  def build(self):
    if self.cache.is_empty():
      return self.build_while_caching()
    tree = {}
    for project in self.cache.get('projects'):
      ancestry = self.get_ancestry_names(project["projectId"])
      if self.resolve:
        ancestry = self.resolve_ancestry(ancestry, project)
      tree = self.graft(tree, ancestry, project)
    return tree

  def build_while_caching(self):
 
    request = self.crm_v1.projects().list()
    tree = {}
    projects_to_cache = []
    while request:
      response = request.execute()
      for project in response.get('projects', []):
        projects_to_cache.append(project)
        ancestry = self.get_ancestry_names(project["projectId"])
        if self.org not in ancestry:
          continue
        if self.resolve:
          ancestry = self.resolve_ancestry(ancestry, project)
        tree = self.graft(tree, ancestry, project)
      request = self.crm_v1.projects().list_next(previous_request=request, previous_response=response)
    self.cache.add("projects", projects_to_cache)
    self.cache.write()
    return tree
  
  def graft(self, tree, ancestry, metadata):
    branch = self.ancestry_branch(ancestry, metadata)
    return merge_or_raise.merge(tree, branch)
  
  def ancestry_branch(self, ancestry, metadata):
    res = {ancestry.pop(): metadata}
    while len(ancestry) > 0:
      res = {ancestry.pop(): res}
    return res 
  
  def resolve_ancestry(self, ancestry, project):
    resolved = []
    for name in ancestry:
      if self.cache.has(name):
        resolved.append(self.cache.get(name))
      else:
        if name.startswith("organizations"):
          response = self.crm_v1.organizations().get(name=name).execute()
        elif name.startswith("folders"):
          response = self.crm_v2.folders().get(name=name).execute()
        elif name.startswith("projects"):
          response = {'displayName': project['projectId']}
        self.cache.add(name, response['displayName'])
        resolved.append(response['displayName'])
    return resolved

  def get_ancestry_names(self, project_id):
    cache_key = "ancestry/{}".format(project_id)
    if self.cache.has(cache_key):
      return self.cache.get(cache_key)
    response = self.crm_v1.projects().getAncestry(projectId=project_id).execute()
    names = list(reversed(["{}s/{}".format(i['resourceId']['type'], i['resourceId']['id']) for i in response["ancestor"]]))
    self.cache.add(cache_key, names)
    return names



import json
import os
from datetime import datetime, timedelta
from tempfile import gettempdir

class Cache():
  TIMESTAMP_FORMAT = "%Y-%m-%d-%H-%M"

  def __init__(self):
    self.data = {}
    self.filename = os.path.join(gettempdir(), "gcptree-cache.json")
    if os.path.exists(self.filename):
      with open(self.filename, 'r') as f:
        data = json.load(f)
        self.data = self.updated_data(data)
        
  def updated_data(self, data):
    previous_timestamp = datetime.strptime(data['_timestamp'], self.TIMESTAMP_FORMAT)
    if previous_timestamp > datetime.now() - timedelta(hours=1) :
      return data
    return {}

  def is_empty(self):
    return len(self.data) == 0

  def has(self, key):
    return key in self.data

  def get(self, key):
    if self.has(key):
      return self.data[key]

  def add(self, key, value):
    self.data[key] = value
  
  def timestamp(self):
    return datetime.now().strftime(self.TIMESTAMP_FORMAT)
  
  def write(self):
    self.data['_timestamp'] = self.timestamp()
    with open(self.filename, 'w') as f:
      json.dump(self.data, f)



import pytest
from datetime import datetime, timedelta
from .cache import Cache

def test_cache_loaded():
  inst = Cache()
  a_few_minutes_ago = datetime.now() - timedelta(minutes=15)
  data = {"_timestamp": a_few_minutes_ago.strftime(inst.TIMESTAMP_FORMAT), "key":"value"}
  assert len(inst.updated_data(data)) > 0

def test_cache_busted():
  inst = Cache()
  over_an_hour_ago = datetime.now() - timedelta(minutes=90)
  data = {"_timestamp": over_an_hour_ago.strftime(inst.TIMESTAMP_FORMAT), "key":"value"}
  assert len(inst.updated_data(data)) == 0


from setuptools import setup
# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
  long_description = f.read()

setup(name='gcptree',
      version='0.1.4',
      description='List your GCP Org heirachy as a tree in JSON or Text',
      url='http://github.com/onetwopunch/gcptree',
      scripts=['bin/gcptree'],
      author='Ryan Canty',
      author_email='onetwopunch@pm.me',
      license='MIT',
      packages=['gcptree'],
      zip_safe=False,
      long_description=long_description,
      long_description_content_type='text/markdown',
      install_requires=[
        "deepmerge==0.2.1",
        "google-api-python-client==1.12.8",
        "colorama==0.4.3",
      ])